
;====================================================================================================

struc MD5_CONTEXT
{
  .lo rd 1
  .hi rd 1
  .hash rd 4
  .wbuf rd 16
}

virtual at edi
  MD5_CTX MD5_CONTEXT
end virtual



macro MD5ASM_INIT
{

  define a eax
  define b ebx
  define c ecx
  define d edx
  define v esi

  macro MD5ASM_STEP func, W, X, Y, Z, idx, t, shift
  \{
    if func = "F"
      ;v = (Z xor (X and (Y xor Z)))
      mov v, Y
      xor v, Z
      and v, X
      xor v, Z
    else if func = "G"
      ;v = (Y xor (Z and (X xor Y)))
      mov v, X
      xor v, Y
      and v, Z
      xor v, Y
    else if func = "H"
      ;v = (X xor Y xor Z)
      mov v, X
      xor v, Y
      xor v, Z
    else if func = "I"
      ;v = (Y xor (X or (not Z)))
      mov v, Z
      not v
      or  v, X
      xor v, Y
    end if
    
    ; W = X + ((W + v + wbuf[idx] + t) <<< shift)
    add v, [MD5_CTX.wbuf+idx*4]
    lea v, [v+W+t]
    rol v, shift
    lea W, [v+X]
  \}



  macro MD5ASM_END
  \{
    restore a
    restore b
    restore c
    restore d
    restore v

    purge MD5ASM_INIT
    purge MD5ASM_STEP
    purge MD5ASM_END
  \}
}


;====================================================================================================

MD5:



proc .body uses ecx edx ebx esi edi, context:DWORD

  mov edi, [context]

  MD5ASM_INIT

  ; copy hash from context to A,B,C,D
  mov a, [MD5_CTX.hash +  0]
  mov b, [MD5_CTX.hash +  4]
  mov c, [MD5_CTX.hash +  8]
  mov d, [MD5_CTX.hash + 12]

  ; Round 1
  MD5ASM_STEP "F", a, b, c, d,  0, 0xD76AA478,  7
  MD5ASM_STEP "F", d, a, b, c,  1, 0xE8C7B756, 12
  MD5ASM_STEP "F", c, d, a, b,  2, 0x242070DB, 17
  MD5ASM_STEP "F", b, c, d, a,  3, 0xC1BDCEEE, 22
  MD5ASM_STEP "F", a, b, c, d,  4, 0xF57C0FAF,  7
  MD5ASM_STEP "F", d, a, b, c,  5, 0x4787C62A, 12
  MD5ASM_STEP "F", c, d, a, b,  6, 0xA8304613, 17
  MD5ASM_STEP "F", b, c, d, a,  7, 0xFD469501, 22
  MD5ASM_STEP "F", a, b, c, d,  8, 0x698098D8,  7
  MD5ASM_STEP "F", d, a, b, c,  9, 0x8B44F7AF, 12
  MD5ASM_STEP "F", c, d, a, b, 10, 0xFFFF5BB1, 17
  MD5ASM_STEP "F", b, c, d, a, 11, 0x895CD7BE, 22
  MD5ASM_STEP "F", a, b, c, d, 12, 0x6B901122,  7
  MD5ASM_STEP "F", d, a, b, c, 13, 0xFD987193, 12
  MD5ASM_STEP "F", c, d, a, b, 14, 0xA679438E, 17
  MD5ASM_STEP "F", b, c, d, a, 15, 0x49B40821, 22

  ; Round 2
  MD5ASM_STEP "G", a, b, c, d,  1, 0xF61E2562,  5
  MD5ASM_STEP "G", d, a, b, c,  6, 0xC040B340,  9
  MD5ASM_STEP "G", c, d, a, b, 11, 0x265E5A51, 14
  MD5ASM_STEP "G", b, c, d, a,  0, 0xE9B6C7AA, 20
  MD5ASM_STEP "G", a, b, c, d,  5, 0xD62F105D,  5
  MD5ASM_STEP "G", d, a, b, c, 10, 0x02441453,  9
  MD5ASM_STEP "G", c, d, a, b, 15, 0xD8A1E681, 14
  MD5ASM_STEP "G", b, c, d, a,  4, 0xE7D3FBC8, 20
  MD5ASM_STEP "G", a, b, c, d,  9, 0x21E1CDE6,  5
  MD5ASM_STEP "G", d, a, b, c, 14, 0xC33707D6,  9
  MD5ASM_STEP "G", c, d, a, b,  3, 0xF4D50D87, 14
  MD5ASM_STEP "G", b, c, d, a,  8, 0x455A14ED, 20
  MD5ASM_STEP "G", a, b, c, d, 13, 0xA9E3E905,  5
  MD5ASM_STEP "G", d, a, b, c,  2, 0xFCEFA3F8,  9
  MD5ASM_STEP "G", c, d, a, b,  7, 0x676F02D9, 14
  MD5ASM_STEP "G", b, c, d, a, 12, 0x8D2A4C8A, 20

  ; Round 3
  MD5ASM_STEP "H", a, b, c, d,  5, 0xFFFA3942,  4
  MD5ASM_STEP "H", d, a, b, c,  8, 0x8771F681, 11
  MD5ASM_STEP "H", c, d, a, b, 11, 0x6D9D6122, 16
  MD5ASM_STEP "H", b, c, d, a, 14, 0xFDE5380C, 23
  MD5ASM_STEP "H", a, b, c, d,  1, 0xA4BEEA44,  4
  MD5ASM_STEP "H", d, a, b, c,  4, 0x4BDECFA9, 11
  MD5ASM_STEP "H", c, d, a, b,  7, 0xF6BB4B60, 16
  MD5ASM_STEP "H", b, c, d, a, 10, 0xBEBFBC70, 23
  MD5ASM_STEP "H", a, b, c, d, 13, 0x289B7EC6,  4
  MD5ASM_STEP "H", d, a, b, c,  0, 0xEAA127FA, 11
  MD5ASM_STEP "H", c, d, a, b,  3, 0xD4EF3085, 16
  MD5ASM_STEP "H", b, c, d, a,  6, 0x04881D05, 23
  MD5ASM_STEP "H", a, b, c, d,  9, 0xD9D4D039,  4
  MD5ASM_STEP "H", d, a, b, c, 12, 0xE6DB99E5, 11
  MD5ASM_STEP "H", c, d, a, b, 15, 0x1FA27CF8, 16
  MD5ASM_STEP "H", b, c, d, a,  2, 0xC4AC5665, 23

  ; Round 4
  MD5ASM_STEP "I", a, b, c, d,  0, 0xF4292244,  6
  MD5ASM_STEP "I", d, a, b, c,  7, 0x432AFF97, 10
  MD5ASM_STEP "I", c, d, a, b, 14, 0xAB9423A7, 15
  MD5ASM_STEP "I", b, c, d, a,  5, 0xFC93A039, 21
  MD5ASM_STEP "I", a, b, c, d, 12, 0x655B59C3,  6
  MD5ASM_STEP "I", d, a, b, c,  3, 0x8F0CCC92, 10
  MD5ASM_STEP "I", c, d, a, b, 10, 0xFFEFF47D, 15
  MD5ASM_STEP "I", b, c, d, a,  1, 0x85845DD1, 21
  MD5ASM_STEP "I", a, b, c, d,  8, 0x6FA87E4F,  6
  MD5ASM_STEP "I", d, a, b, c, 15, 0xFE2CE6E0, 10
  MD5ASM_STEP "I", c, d, a, b,  6, 0xA3014314, 15
  MD5ASM_STEP "I", b, c, d, a, 13, 0x4E0811A1, 21
  MD5ASM_STEP "I", a, b, c, d,  4, 0xF7537E82,  6
  MD5ASM_STEP "I", d, a, b, c, 11, 0xBD3AF235, 10
  MD5ASM_STEP "I", c, d, a, b,  2, 0x2AD7D2BB, 15
  MD5ASM_STEP "I", b, c, d, a,  9, 0xEB86D391, 21

  ; store hash back to context
  add [MD5_CTX.hash +  0], a
  add [MD5_CTX.hash +  4], b
  add [MD5_CTX.hash +  8], c
  add [MD5_CTX.hash + 12], d

  MD5ASM_END

  ret
endp



proc .init uses edi, context:DWORD

  mov edi, [context]

  mov [MD5_CTX.lo], 0
  mov [MD5_CTX.hi], 0

  mov [MD5_CTX.hash +  0], 0x67452301
  mov [MD5_CTX.hash +  4], 0xEFCDAB89
  mov [MD5_CTX.hash +  8], 0x98BADCFE
  mov [MD5_CTX.hash + 12], 0x10325476

  ret
endp



proc .update uses ecx edx ebx esi edi, context:DWORD, data:DWORD, datalen:DWORD
  local\
    dataptr:DWORD

  mov edi, [context]

  mov eax, [data]
  mov [dataptr], eax ; pointer to block of data

  mov eax, [MD5_CTX.lo]
  and eax, 0x3F
  mov esi, eax  ; used in wbuf

  lea ebx, [esi - 64]
  neg ebx  ; available in wbuf

  mov eax, [datalen]
  add [MD5_CTX.lo], eax
  adc [MD5_CTX.hi], 0

  mov ecx, eax  ; data bytes left to process

  .while ecx
    .if signed ecx > ebx
      stdcall memcpy, addr MD5_CTX.wbuf+esi, [dataptr], ebx
      stdcall MD5.body, edi
      add [dataptr], ebx
      sub ecx, ebx
      xor esi, esi
      mov ebx, 64
    .else
      stdcall memcpy, addr MD5_CTX.wbuf+esi, [dataptr], ecx
      xor ecx, ecx
    .endif
  .endw

  ret
endp



proc .final uses ebx edx esi edi, context:DWORD, hash:DWORD

  mov edi, [context]

  mov eax, [MD5_CTX.lo]
  and eax, 0x3F
  mov esi, eax  ; used in wbuf

  lea ebx, [esi - 64]
  neg ebx  ; available in wbuf

  mov byte [MD5_CTX.wbuf+esi], 0x80
  inc esi
  dec ebx
  stdcall memset, addr MD5_CTX.wbuf+esi, 0, ebx

  .if esi > 56
    stdcall MD5.body, edi
    stdcall memset, addr MD5_CTX.wbuf, 0, 56
  .endif

  ; write total length of hashed data, in bits
  mov eax, [MD5_CTX.lo]
  mov edx, eax
  shr edx, 29
  shl eax, 3
  mov [MD5_CTX.wbuf+14*4], eax
  mov eax, [MD5_CTX.hi]
  shl eax, 3
  or eax, edx
  mov [MD5_CTX.wbuf+15*4], eax

  stdcall MD5.body, edi

  stdcall memcpy, [hash], addr MD5_CTX.hash, 16

  ret
endp



proc .calc uses edi, data:DWORD, datalen:DWORD, hash:DWORD
  local \
    context:MD5_CONTEXT

  lea edi, [context]
  stdcall MD5.init, edi
  stdcall MD5.update, edi, [data], [datalen]
  stdcall MD5.final, edi, [hash]

  ret
endp



PURGE_MD5_NAMESPACE:
